// 音声メッセージ付電子オルゴール MyMelo2PLUS(録音音声の再生機能を追加)
// https://koyama.verse.jp/elecraft/mymelo
// 20230604 koyama88@cameo.plala.or.jp
// (特徴)
// (1) MyMelo2 Plus は、私だけの電子オルゴール MyMelo2 に録音音声を挿入する機能を追加したもので、「音声メッセージ付 電子オルゴール」を作ることができます。
// (2) 挿入できる音声メッセージは、中品質で最大1.3秒、低品質で最大2.6秒程度です。
// (3) 音声メッセージはオルゴールのメロディーの任意の場所に挿入できます。複数の音声を別々の場所で再生することもできます。
// (4) 使用するマイコンはATtiny85です。Digisparkでも動作します。
//
// (音声データの作成)
// (1) Audacityなどのソフトを使い、パソコンで音声を録音し、wavファイルとして保存します。スマホで録音したファイルをFFMPEGなどのソフトを使って変換することもできます。
// (2) 録音音声ファイル(〇〇.wav)を mymelo2plus.wsf にドラッグ＆ドロップすると、mymelo2plusフォルダの中に音声データファイル myvoice.h が作られます。
// (3) 複数の音声ファイル(〇〇.wav)を用意し、(2)の操作を繰り返すと、音声データファイルに複数の音声を登録できます。ただし、記録できる時間は「合計で」最大1.3秒程度(中品質)または2.6秒程度(低品質)です。
//
// (楽譜データの作成)
// (1) 「テキスト音楽サクラ」(またはメモ帳などのテキストエディタ)で「ドドソソララソー」のように楽譜ファイル(〇〇.mml)を作成します。
// (2) オルゴールの音はmymelo2の「減衰正弦波」と同じです。
// (3) オルゴールの楽譜データの作り方はmymelo2と同じで、2和音や3和音も可能です。
// (4) 録音音声の箇所はmmlの楽譜データの中に「n0」「n1」...のように記載します。
// (5) 楽譜ファイル(〇〇.mml)をmymelo2plus.wsfにドラッグ＆ドロップすると、mymelo2plusフォルダの中に楽譜データファイル(mymelo2.h)が作られます。
// (6) 複数の楽譜ファイル(〇〇.mml)を作成し、(5)の操作を繰り返すと、楽譜データファイルに複数(2つまで)の楽譜を登録できます。
//
// (プログラムの書き込み)
// ArduinoIDEを起動し、mymelo2plusフォルダの mymelo2plus.ino を開いて書き込みます。
//
// (動作)
// 電源を投入すると、録音メッセージ付のオルゴールが再生されます。外部スイッチをつなぐと、スイッチを押したときに再生されます。スイッチを長押しすると2曲目が再生されます。

#include <avr/sleep.h>
#include "mymelo2.h"						// 楽譜データ
#include "myvoice.h"						// 録音音声データ
#define btnApin 2
#define Bzz1pin	1
#define Bzz2pin	4
#define _Voice(i,p)	pgm_read_byte(&((char*)pgm_read_ptr(&Voice[i]))[p])
#define _len(i,n)	pgm_read_byte(&((uint8_t*)pgm_read_ptr(&pLen[i]))[n])
#define _Part1(i,p)	pgm_read_byte(&((uint8_t*)pgm_read_ptr(&pPart1[i]))[p])
#define _Part2(i,p)	pgm_read_byte(&((uint8_t*)pgm_read_ptr(&pPart2[i]))[p])
#define _Part3(i,p)	pgm_read_byte(&((uint8_t*)pgm_read_ptr(&pPart3[i]))[p])
volatile uint8_t btnAcntr=0;		// チャタリング防止カウンタ 32*255=8.1ms
volatile uint16_t btnAcntr2=0;	// 長押し用カウンタ(volatile必要)
int8_t	nPlay=1;								// 演奏回数
#if defined(MYVOICE)	// -------------------------------------------------------
// Original Code for PIC by Rodger Richey, 1-9-96
// This ADPCM routines were obtained from the Interactive Multimedia Association's
// Reference ADPCM algorithm.  This algorithm was first implemented by Intel/DVI.
#if CODESIZE==4				// -------------------------------------------------------
char IndexTable[]={-1,-1,-1,-1,2,4,6,8};	// Table of index changes
#elif CODESIZE==2			// -------------------------------------------------------
char IndexTable[]={-1,2};									// Table of index changes
#endif								// -------------------------------------------------------
const int16_t StepSizeTable[] PROGMEM={		// Quantizer step size lookup table
	7, 8, 9, 10, 11, 12, 13, 14, 16, 17,
	19, 21, 23, 25, 28, 31, 34, 37, 41, 45,
	50, 55, 60, 66, 73, 80, 88, 97, 107, 118,
	130, 143, 157, 173, 190, 209, 230, 253, 279, 307,
	337, 371, 408, 449, 494, 544, 598, 658, 724, 796,
	876, 963, 1060, 1166, 1282, 1411, 1552, 1707, 1878, 2066,
	2272, 2499, 2749, 3024, 3327, 3660, 4026, 4428, 4871, 5358,
	5894, 6484, 7132, 7845, 8630, 9493, 10442, 11487, 12635, 13899,
	15289, 16818, 18500, 20350, 22385, 24623, 27086, 29794, 32767
};
int16_t lastsample=0;									// Output of ADPCM predictor
int8_t lastindex=0;										// Quantizer step size index
uint8_t code;													// dual 4-bits / quad 2-bits ADPCM code
uint8_t j=0;													// upper 4-bit / lower 4-bit of code
uint8_t ivoice=0;											// Voice[0]を再生
uint16_t p=voicesize[ivoice];					// 音声データのポインタ(初期値は終了状態)

int16_t ADPCMDecoder(uint8_t code){		// code: dual 4-bits / quad 2-bits ADPCM code
	int16_t predsample=lastsample;			// Restore previous values of predicted sample
	int8_t index=lastindex;							// and quantizer step size index
	int16_t diffq;											// Dequantized predicted difference
// Find quantizer step size from lookup table using index
	int16_t step=pgm_read_word(&StepSizeTable[index]);	// Quantizer step size
// Inverse quantize the ADPCM code into a difference using the quantizer step size
#if CODESIZE==4				// -------------------------------------------------------
	diffq=step>>3;
	if(code&4)	diffq+=step;
	if(code&2)	diffq+=step>>1;
	if(code&1)	diffq+=step>>2;
// Add the difference to the predicted sample
	if(code&8){
#elif CODESIZE==2			// -------------------------------------------------------
	diffq=step>>1;
	if(code&1)	diffq+=step;
// Add the difference to the predicted sample
	if(code&2){
#endif								// -------------------------------------------------------
// Check for overflow of the new predicted sample
		if(predsample<0 && predsample<diffq-32768)	predsample=-32768;
		else																				predsample-=diffq;
	}else{
		if(predsample>=0 && diffq>32767-predsample)	predsample=32767;
		else																				predsample+=diffq;
	}
// Find new quantizer step size by adding the old index and a table lookup
// using the ADPCM code
#if CODESIZE==4				// -------------------------------------------------------
	index+=IndexTable[code&0x07];
#elif CODESIZE==2			// -------------------------------------------------------
	index+=IndexTable[code&0x01];
#endif								// -------------------------------------------------------
// Check for overflow of index the new quantizer step size
	if(index<0)				index=0;
	else if(index>88)	index=88;
// Save predicted sample for next iteration and quantizer step size index
	lastsample=predsample;	lastindex=index;
	return predsample;									// Return the new speech sample
}
#endif								// -------------------------------------------------------

uint8_t imelo=0;											// 演奏中の曲番号(初期値は0)
uint8_t decayTimer;										// 32μs*256=8.192ms毎に包絡を減衰
uint16_t CNT, cntr;										// 32us*(1953*120/Tempo)=62.5ms
uint16_t p1, p2, p3;									// 楽譜part1～part3のポインタ
uint8_t note1,note2,note3;						// 楽譜データDO|L4
uint8_t len1, len2, len3;							// 音符の長さ
uint16_t v,v1,v2,v3, u,u1,u2,u3;			// 振幅とその合算値

#define _Ws(x)	(uint16_t)pgm_read_byte(&Ws[(uint8_t)(x)])
#define _SW(i,n)	pgm_read_word(&((uint16_t*)pgm_read_ptr(&pSW[i]))[n])
uint8_t len1a=0,len2a=0,len3a=0;			// アタックタイム 32μs*128=4.096ms
uint16_t E,E1,E2,E3;									// 包絡
uint16_t pW1,pW2,pW3, sw1,sw2,sw3;		// 波形ポインタとサンプリングステップ幅
uint8_t x,y;
// 正弦波の波形テーブル(値:0～255, データ数:256)
const uint8_t Ws[] PROGMEM={0,0,0,0,1,1,1,2,2,3,4,5,5,6,7,9,10,11,12,14,15,17,18,20,21,23,25,27,29,31,33,35,37,40,42,44,47,49,52,54,57,59,62,65,67,70,73,76,79,82,85,88,90,93,97,100,103,106,109,112,115,118,121,124,127,131,134,137,140,143,146,149,152,155,158,162,165,167,170,173,176,179,182,185,188,190,193,196,198,201,203,206,208,211,213,215,218,220,222,224,226,228,230,232,234,235,237,238,240,241,243,244,245,246,248,249,250,250,251,252,253,253,254,254,254,255,255,255,255,255,255,255,254,254,254,253,253,252,251,250,250,249,248,246,245,244,243,241,240,238,237,235,234,232,230,228,226,224,222,220,218,215,213,211,208,206,203,201,198,196,193,190,188,185,182,179,176,173,170,167,165,162,158,155,152,149,146,143,140,137,134,131,128,124,121,118,115,112,109,106,103,100,97,93,90,88,85,82,79,76,73,70,67,65,62,59,57,54,52,49,47,44,42,40,37,35,33,31,29,27,25,23,21,20,18,17,15,14,12,11,10,9,7,6,5,5,4,3,2,2,1,1,1,0,0,0};

ISR(TIMER0_COMPA_vect){								// 32us毎/125us毎に1サンプル出力
	if(digitalRead(btnApin)==LOW)	btnAcntr=255;	// btnAが押されていれば
	else if(btnAcntr>0)						btnAcntr--;		// 32*255=8.1msのチャタリング防止
	if(btnAcntr2>0)								btnAcntr2--;	// 長押し判別カウンタ

	if(nPlay>0){												// 演奏回数
#if defined(MYVOICE)	// --- MYVOICE -------------------------------------------
		if(p<voicesize[ivoice]){					// 音声再生中なら ************************
			int16_t val;
#if CODESIZE==4				// -------------------------------------------------------
			if(j==0){
				code=_Voice(ivoice, p);
				val=ADPCMDecoder(code>>4);		// upper 4-bits -> 16-bits
 			}else{
				val=ADPCMDecoder(code);				// lower 4-bits -> 16-bits
				if(++p==voicesize[ivoice]){		// 再生終了なら
					OCR0A=32*F_CPU/8/1000000;		// TC0の割込み周期..0.5us*64=32us
					cntr=1;	len1=len2=len3=1;		// すぐに次のデータ
				}
			}
			j++;	j&=0x01;
#elif CODESIZE==2			// -------------------------------------------------------
			if(j==0){
				code=_Voice(ivoice, p);
				val=ADPCMDecoder(code>>6);		// bit7-6 -> 16-bits
			}else if(j==1){
				val=ADPCMDecoder(code>>4);		// bit5-4 -> 16-bits
 			}else if(j==2){
				val=ADPCMDecoder(code>>2);		// bit3-2 -> 16-bits
 			}else{
				val=ADPCMDecoder(code);				// bit1-0 -> 16-bits
				if(++p==voicesize[ivoice]){		// 再生終了なら
					OCR0A=32*F_CPU/8/1000000;		// TC0の割込み周期..0.5us*64=32us
					cntr=1;	len1=len2=len3=1;		// すぐに次のデータ
				}
		 	}
			j++;	j&=0x03;
#endif								// --- CODESIZE ------------------------------------------
			OCR1A=128+(val>>8);	OCR1B=128-(val>>8);
		}else{	// 音声再生中でなければ ********************************************
#endif								// --- MYVOICE -------------------------------------------

			if(--cntr==0){											// 32us*(1953*120/Tempo)=62.5ms
				cntr=CNT;
				if(--len1==0){
					note1=_Part1(imelo, p1++);	len1=_len(imelo, note1>>5);	note1&=0x1f;
					if(note1==0)	nPlay=0;					// part1終了!!
#if defined(MYVOICE)	// --- MYVOICE -------------------------------------------
					else if(note1<=NVOICE){					// 音声再生なら
						OCR0A=250;										// TC0の割込み周期..0.5us*250=125us
						ivoice=note1-1;								// 再生する音声番号 0～
						p=0; lastsample=0; lastindex=0;	// 音声サンプル先頭から
					}
#endif								// -------------------------------------------------------
					else{														// sw1: 波形テーブルのステップ幅
						sw1=_SW(imelo, note1);	E1=0;	// E1: 初めは0 32*64=2.048ms後に最大0x8000
					}
					len1a=128;											// アタックタイム(32*128=4.096ms)
				}

#if NTRACK==1					// --- TRACK数1 ------------------------------------------
			}
			u1=0;	v1=v3=0;
			if(note1==NVOICE+1){	v1=v3=255;	u1=127;	}	// 休符
			else if(note1>NVOICE+1){						// part1の振幅を加算
				if(len1a>0){	len1a--;	E1+=256;}	// アタックタイム後にE1=0x8000
				E=0x8000-E1;
				u1=_Ws(E>>9);	u1-=(u1>>5);				// 0-96
				x=(pW1+=sw1)>>8;	y=(pW1+E)>>8;	v1=_Ws(x)+_Ws(y);	// 0-510
				x=(pW3+=(sw1+10))>>8;	y=(pW3+E)>>8;	// わずか周波数をずらす
				v3=_Ws(x)+_Ws(y);									// 0-510
			}
			v=v1+((v1+v3)>>1);									// 0-1020
			u=(u1<<2);													// 0-384
			OCR1A=(v-u)>>2;	OCR1B=(1024-v-u)>>2;// 0-255
			if(--decayTimer==0)	E1-=(E1>>5);		// 32us*256=8.192ms毎に減衰(31/32)

#elif NTRACK==2				// --- TRACK数2 ----------------------------------------
				if(--len2==0){
					note2=_Part2(imelo, p2++);	len2=_len(imelo, note2>>5);	note2&=0x1f;
					if(note2==0)	p2--;							// Part2終了!!
					else if(note2<=NVOICE){	}				// 音声再生なら何もしない
					else{														// sw2: 波形テーブルのステップ幅
						sw2=_SW(imelo, note2);	E2=0;	// E2: 初めは0 32*64=2.048ms後に最大0x8000
					}
					len2a=128;											// アタックタイム 32*128=4.096ms
				}
			}
			u1=u2=0;	v1=v2=v3=0;
			if(note1==NVOICE+1){	v1=v3=255;	u1=127;	}	// 休符
			else if(note1>NVOICE+1){								// part1の振幅を加算
				if(len1a>0){	len1a--;	E1+=256;	}		// アタックタイム後にE1=0x8000
				E=0x8000-E1;			u1=_Ws(E>>9);	u1-=(u1>>5);			// 0-96
				x=(pW1+=sw1)>>8;	y=(pW1+E)>>8;	v1=_Ws(x)+_Ws(y);	// 0-510
				x=(pW3+=(sw1+10))>>8;	y=(pW3+E)>>8;		// わずか周波数をずらす
				v3=_Ws(x)+_Ws(y);																	// 0-510
			}
			if(note2==NVOICE+1){	v2=255;	u2=127;	}	// 休符
			else if(note2>NVOICE+1){								// part2の振幅を加算
				if(len2a>0){	len2a--;	E2+=256;	}		// アタックタイム後にE2=0x8000
				E=0x8000-E2;			u2=_Ws(E>>9);	u2-=(u2>>5);			// 0-96
				x=(pW2+=sw2)>>8;	y=(pW2+E)>>8;	v2=_Ws(x)+_Ws(y);	// 0-510
			}
			v=v1+((v2+v3)>>1)+2;										// 0-1020
			u=(u1<<1)+u2+u1;												// 0-384
			OCR1A=(v-u)>>2;	OCR1B=(1024-v-u)>>2;		// 0-255
			if(--decayTimer==0){										// 32us*256=8.192ms毎に減衰(31/32)
				E1-=(E1>>5); E2-=(E2>>5);
			}

#elif NTRACK==3				// --- TRACK数3 ------------------------------------------
				if(--len2==0){
					note2=_Part2(imelo, p2++);	len2=_len(imelo, note2>>5);	note2&=0x1f;
					if(note2==0)	p2--;							// Part2終了!!
					else if(note2<=NVOICE){	}				// 音声再生なら何もしない
					else{														// sw2: 波形テーブルのステップ幅
						sw2=_SW(imelo, note2);	E2=0;	// E2: 初めは0 32*128=4.096ms後に最大0x8000
					}
					len2a=128;											// アタックタイム 32*128=4.096ms
				}
				if(--len3==0){
					note3=_Part3(imelo, p3++);	len3=_len(imelo, note3>>5);	note3&=0x1f;
					if(note3==0)	p3--;							// Part3終了!!
					else if(note3<=NVOICE){	}				// 音声再生なら何もしない
					else{														// sw3: 波形テーブルのステップ幅
						sw3=_SW(imelo, note3);	E3=0;	// E3: 初めは0 32*128=4.096ms後に最大0x8000
					}
					len3a=128;											// アタックタイム 32*128=4.096ms
				}
			}
			u1=u2=u3=0;	v1=v2=v3=0;
			if(note1==NVOICE+1){	v1=255;	u1=127;	}	// 休符
			else if(note1>NVOICE+1){								// part1の振幅を加算
				if(len1a>0){	len1a--;	E1+=256;	}		// アタックタイム後にE1=0x8000
				E=0x8000-E1;			u1=_Ws(E>>9);	u1-=(u1>>5);			// 0-96
				x=(pW1+=sw1)>>8;	y=(pW1+E)>>8;	v1=_Ws(x)+_Ws(y);	// 0-510
			}
			if(note2==NVOICE+1){	v2=255;	u2=127;	}	// 休符
			else if(note2>NVOICE+1){								// part2の振幅を加算
				if(len2a>0){	len2a--;	E2+=256;	}		// アタックタイム後にE2=0x8000
				E=0x8000-E2;			u2=_Ws(E>>9);	u2-=(u2>>5);			// 0-96
				x=(pW2+=sw2)>>8;	y=(pW2+E)>>8;	v2=_Ws(x)+_Ws(y);	// 0-510
			}
			if(note3==NVOICE+1){	v3=255;	u3=127;	}	// 休符
			else if(note3>NVOICE+1){								// part3の振幅を加算
				if(len3a>0){	len3a--;	E3+=256;	}		// アタックタイム後にE3=0x8000
				E=0x8000-E3;			u3=_Ws(E>>9);	u3-=(u3>>5);			// 0-96
				x=(pW3+=sw3)>>8;	y=(pW3+E)>>8;	v3=_Ws(x)+_Ws(y);	// 0-510
			}
			v=v1+((v2+v3)>>1)+2;										// 0-1020
			u=(u1<<1)+u2+u3;												// 0-384
			OCR1A=(v-u)>>2;	OCR1B=(1024-v-u)>>2;		// 0-255
			if(--decayTimer==0){										// 32us*256=8.192ms毎に減衰(31/32)
				E1-=(E1>>5); E2-=(E2>>5); E3-=(E3>>5);
			}
#endif								// -------------------------------------------------------
#if defined(MYVOICE)	// --- MYVOICE -------------------------------------------
#if defined(__TinyX5__)			// -------------------------------------------------
			OCR0A=32*F_CPU/8/1000000;				// TC0の割込み周期..0.5us*64=32us
#endif											// -------------------------------------------------
		}									// 		if(p<voicesize[ivoice]){}else{	// 音声再生中でなければ
#else									// -------------------------------------------------------
		OCR0A=32*F_CPU/8/1000000;					// TC0の割込み周期..0.5us*64=32us
#endif								// -------------------------------------------------------
	}										// if(nplay>0){
}

ISR(INT0_vect){}											// btnAオンでwakeup

void setup(){
	pinMode(btnApin, INPUT_PULLUP);	pinMode(Bzz1pin, OUTPUT);	pinMode(Bzz2pin, OUTPUT);
	PLLCSR=0x06;												// TC1...250kHz PWM
	TCCR1=0x51;													// if(TCNT1==1)						OC1A=1;
																			// else if(TCNT1==OCR1A)	OC1A=0;
	GTCCR=0x50;													// if(TCNT1==1)						OC1B=1;
																			// else if(TCNT1==OCR1B)	OC1B=0;
	OCR1C=255;													// fPWM=64000/(255+1)=250kHz
	OCR0A=32*F_CPU/8/1000000;						// TC0の割込み周期..0.5us*64=32us
	TCCR0A=0x02;	TCCR0B=0x02;					// CTC, F_CPU/8=2MHz(0.5us)
	TIMSK=(1<<OCIE0A);	GIMSK=_BV(INT0);	sei();// Int0(外部スイッチ)を有効にする
	beginMelo(0);												// 電源投入時に1曲目を1回演奏
}
void beginMelo(uint8_t i){						// 電子オルゴールi曲目を1回演奏
	OCR0A=32*F_CPU/8/1000000;						// TC0の割込み周期..0.5us*64=32us
	imelo=(i<NMELO?i:0);	nPlay=1;			// i曲目を1回演奏
	CNT=8000000/256*0.5/8*120/tempo[imelo];	// テンポ120の四分音符(長さ8)は0.5秒
	cntr=CNT;														// 32us*(1953*120/TEMPO)=62.5ms
	p1=p2=p3=0;	pW1=pW2=pW3=0;					// ポインタ先頭
	note1=note2=note3=0;	len1=len2=len3=8;	// 初期ディレイ 0.5秒(無音)
}
void loop(){
	if(btnAcntr>0){											// btnAがオンなら
		for(btnAcntr2=15625;btnAcntr>0;);	// オフを待って
		if(nPlay>0)	nPlay=0;							// 演奏中ならば停止
		else{
			if(btnAcntr2>0)	beginMelo(0);		// btnA短押し(32us*15625=0.5s未満)は1曲目
			else						beginMelo(1);		// btnA長押しは2曲目
		}
	}else if(nPlay==0){									// ボタンが押されてなくて再生終了ならスリープ
		pinMode(Bzz1pin, INPUT);	pinMode(Bzz2pin, INPUT);
		GIMSK=_BV(INT0);									// SLEEP中のINT0(btnApin)割込み許可
		set_sleep_mode(SLEEP_MODE_PWR_DOWN);
		sleep_mode();											// SE=1, sleep, SE=0
		GIMSK=0;													// INT0(btnApin)割込み禁止
		pinMode(Bzz1pin, OUTPUT);	pinMode(Bzz2pin, OUTPUT);
	}
}